# module containing the Player class and its functions

import time
from cetragm.tables import thresholds, gravity

class Player:
    def __init__(self):
        self.score = 0
        self.time_ms = 0
        self.start_time = time.perf_counter()
        self.grade = "9"
        self.level = 0
        self.line_goal = 100
        self.hold_piece = ""
        self.active_piece = {}
        
        # checks for midgame
        self.can_gm = True
        self.met_gm_condition_300 = False
        self.met_gm_condition_500 = False
        self.met_gm_condition_999 = False
        
        self.combo = 0
        self.soft = 0
        self.hold_lock = False
        self.lock_resets = 0
        self.fall_progress = 0.0
    
    def upd_time(self):
        self.time_ms = int((time.perf_counter() - self.start_time) * 1000)    

    def check_grade(self):
        self.upd_time()
        # checking for GM eligibility (called every piece, so it'll be up to date)
        if not self.met_gm_condition_300 and self.level >= 300:
            if self.time_ms <= 225000 and self.score >= 12000 and self.can_gm: # 4m 15s 
                self.met_gm_condition_300 = True
            else:
                self.can_gm = False
                
        if not self.met_gm_condition_500 and self.level >= 500:
            if self.time_ms <= 450000 and self.score >= 40000 and self.can_gm: # 7m 30s 
                self.met_gm_condition_500 = True
            else:
                self.can_gm = False
            
        for name, val in thresholds.items(): # actually set grade
            if self.score >= val:
                if not self.grade == "Gm":
                    self.grade = name
            else:
                break
    
    def check_gm(self):
        self.upd_time()
        if not self.can_gm:
            return False
        
        if not (self.met_gm_condition_300 and self.met_gm_condition_500):
            return False
        
        if self.time_ms <= 810000 and self.score >= 126000:
            self.met_gm_condition_999 = True
            self.grade = "Gm"
            return True        
        return False
    
    def get_grav(self):
        keys = sorted(gravity.keys())
        current_g = gravity[0]
        for k in keys:
            if self.level >= k:
                current_g = gravity[k]
            else: 
                break
            
        if current_g <= 0:
            return float('inf')
        return current_g


# Main module
import os
import sys
import time
import threading
import signal
import queue

from cetragm.draw import draw_board
from cetragm.game import lock_piece, collides
from cetragm.player import Player
from cetragm.bag import Bag
from cetragm.controls import InputHandler
from cetragm.srs import rotate_srs
from cetragm.tables import pieces

def sigint_handler(sig, frame):
    sigint.set()
    print("\x1b[2j\x1b[Hgoodbye!")

sigint = threading.Event()
lose = threading.Event()
board_lock = threading.Lock()

signal.signal(signal.SIGINT, sigint_handler)

def setup_board(rows, cols):
    board = [[[0] for _ in range(cols)] for _ in range(rows)]
    return board

def is_grounded(active_piece, board):
    if not active_piece or "pos" not in active_piece:
        return False
    temp_pos = active_piece["pos"][:]
    temp_pos[1] += 1
    return collides({"name": active_piece["name"], "rotation": active_piece["rotation"], "pos": temp_pos}, board)

def get_minos(active_piece):
    if not active_piece or "name" not in active_piece:
        return set()
    try:
        rot_matrix = pieces[active_piece["name"].lower()]["rotations"][active_piece["rotation"]]
    except KeyError:
        return set()
    minos = set()
    for dy, row in enumerate(rot_matrix):
        for dx, cell in enumerate(row):
            if cell:
                minos.add((active_piece["pos"][0] + dx, active_piece["pos"][1] + dy))
    return minos
    
def lock_and_are(player, board, shared):
    board, cleared, loss = lock_piece(player.active_piece, board, player)
    shared["board"] = board
    
    if loss:
        lose.set()
        return "loss", 0.0
    
    player.active_piece = {}
    player.fall_progress = 0.0
    player.hold_lock = False
    player.soft = 0
    player.lock_resets = 0
    
    if cleared:
        return "line_clear", 0.0
    else:
        return "are", 0.0
    
    
def lock_now(player, board, shared, bag):
    try:
        board, cleared, loss = lock_piece(player.active_piece, board, player)
    except ValueError:
        lose.set()
        return "loss"
    shared["board"] = board
    
    if loss:
        lose.set()
        return "loss"
    
    player.active_piece = {}
    player.fall_progress = 0.0
    player.soft = 0
    player.hold_lock = False
    player.lock_resets = 0
    if cleared:
        return "line_clear"
    else:
        player.active_piece = {"name": bag.get_piece(), "pos": [3, 0], "rotation": "0"}
        return "active"

def input_handler(player, board, action, bag, shared, LOCK_DELAY, FRAME):
    if not player.active_piece:
        return None
    
    piece = player.active_piece
    x, y = piece["pos"]
    
    if action == "pause":
        sigint.set()
        return None

    elif action == "move_left":
        piece["pos"][0] -= 1
        if collides(piece, board):
            piece["pos"][0] += 1
        return None
    
    elif action == "move_right":
        piece["pos"][0] += 1
        if collides(piece, board):
            piece["pos"][0] -= 1
        return None
    
    elif action == "soft_drop":
        piece["pos"][1] += 1
        if collides(piece, board):
            piece["pos"][1] -= 1
            return "ground"
        player.soft += 1
        return None
    
    elif action == "hard_drop":
        while not collides(piece, board):
            piece["pos"][1] += 1
        piece["pos"][1] -= 1
        player.soft += 4
        return lock_now(player, board, shared, bag)
    
    elif action == "rotate_cw":
        rotate_srs(player.active_piece, +1, board, 
            lambda n,r,p: collides({"name": n, "rotation": r, "pos": [p[0], p[1]]}, board))
        return None
            
    elif action == "rotate_ccw":
        rotate_srs(player.active_piece, -1, board, 
            lambda n,r,p: collides({"name": n, "rotation": r, "pos": [p[0], p[1]]}, board))
        return None
            
    elif action == "rotate_180":
        rotate_srs(player.active_piece, 2, board, 
            lambda n,r,p: collides({"name": n, "rotation": r, "pos": [p[0], p[1]]}, board))
        return None
            
    elif action == "hold":
        if player.hold_lock:
            return None
        old = player.hold_piece
        player.hold_piece = player.active_piece["name"]
        if not old:
            player.active_piece = {"name": bag.get_piece(), "pos": [3, 0], "rotation": "0"}
        else:
            player.active_piece = {"name": old, "pos": [3, 0], "rotation": "0"}
            
        player.fall_progress = 0.0
        player.hold_lock = True
        player.lock_resets = 0
        return None
    
    return None
    
def render_loop(shared, player, bag, fps):
    frame_time = 1.0 / fps
    while not sigint.is_set():
        if lose.is_set():
            sigint.set()
        with board_lock:
            board = shared["board"]
        
        start = time.perf_counter()
        player.upd_time()
        print("\x1b[H\x1b[2J")
        draw_board(
            board,
            player.active_piece,
            player.score,
            player.grade,
            player.time_ms,
            player.level,
            player.line_goal,
            player.hold_piece,
            bag.get_preview()
        ) # got damn
        elapsed = time.perf_counter() - start
        sleep_time = frame_time - elapsed
        if sleep_time > 0:
            time.sleep(sleep_time)
        else:
            print("\x1b[H\x1b[31mRunning below 60fps!! Performance will be degraded\x1b[0m")

def game_loop(shared, player, bag, inputs, fps):
    from cetragm.config import ARE_FRAMES, LINE_CLEAR_FRAMES, LOCK_DELAY_FRAMES
    FRAME = 1.0 / fps
    
    LOCK_DELAY = LOCK_DELAY_FRAMES * FRAME
    ARE_DELAY = ARE_FRAMES * FRAME
    LINE_CLEAR_DELAY = LINE_CLEAR_FRAMES * FRAME
    
    state = "spawn"
    lock_timer = 0.0
    phase_timer = 0.0
    
    last_time = time.perf_counter()
    
    if not hasattr(player, "fall_progress"):
        player.fall_progress = 0.0
    
    if not hasattr(player, "lock_resets"):
        player.lock_resets = 0
    
    while not sigint.is_set():
        now = time.perf_counter()
        dt = now - last_time
        last_time = now
        if dt > 0.1:
            dt = 0.1
        
        with board_lock:
            board = shared["board"]
            while True:
                try:
                    action = inputs.queue.get_nowait()
                except queue.Empty:
                    break
                
                was_grounded = is_grounded(player.active_piece, board)
                old_minos = get_minos(player.active_piece)
                
                new_state = input_handler(player, board, action, bag, shared, LOCK_DELAY, FRAME)
                
                if new_state == "ground":
                    lock_timer += dt
                    if lock_timer >= LOCK_DELAY:
                        state, lock_timer = lock_and_are(player, board, shared)
                        if state == "loss":
                            return
                        phase_timer = 0.0
                    continue
                
                if new_state in ("line_clear", "are", "loss", "active"):
                    state = new_state
                    phase_timer = 0.0
                    player.hold_lock = False
                    if state == "active":
                        player.fall_progress = 0.0
                    break
                
                new_minos = get_minos(player.active_piece)
                if new_minos != old_minos and was_grounded:
                    now_grounded = is_grounded(player.active_piece, board)
                    if now_grounded:
                        player.lock_resets += 1
                        if player.lock_resets <= 15:
                            lock_timer = 0.0
                    else:
                        player.lock_resets = 0
            
            player.check_grade()
            
            # gravity
            if state == "active":
                g_units = player.get_grav()
                cells_per_frame = g_units / 256.0
                cells_per_second = cells_per_frame * fps
                player.fall_progress += cells_per_second * dt
                cells_to_fall = int(player.fall_progress)
                
                if cells_to_fall > 0:
                    player.fall_progress -= cells_to_fall
                    piece = player.active_piece
                    
                    for _ in range(cells_to_fall):
                        piece["pos"][1] += 1
                        if collides(piece, board):
                            piece["pos"][1] -= 1
                            lock_timer += dt
                            if lock_timer >= LOCK_DELAY:
                                state, lock_timer = lock_and_are(player, board, shared)
                                if state == "loss":
                                    return
                                phase_timer = 0.0
                            break
                    else:
                        lock_timer = 0.0
                        
                if state == "active":
                    piece = player.active_piece
                    temp_pos = piece["pos"].copy()
                    temp_pos[1] += 1
                    if collides({"name": piece["name"], "pos": temp_pos, "rotation": piece["rotation"]}, board):
                        lock_timer += dt
                        if lock_timer >= LOCK_DELAY:
                            state, lock_timer = lock_and_are(player, board, shared)
                            if state == "loss":
                                return
                            phase_timer = 0.0
            
            # pause after line clear
            elif state == "line_clear":
                phase_timer += dt
                if phase_timer >= LINE_CLEAR_DELAY:
                    state = "are"
                    phase_timer = 0.0
            
            # pause after a piece locking without soft drop
            elif state == "are":
                phase_timer += dt
                if phase_timer >= ARE_DELAY:
                    player.active_piece = {"name": bag.get_piece(), "pos": [3, 0], "rotation": "0"}
                    player.fall_progress = 0.0
                    lock_timer = 0.0
                    player.lock_resets = 0
                    state = "active"
            
            # first piece
            elif state == "spawn":
                player.active_piece = {"name": bag.get_piece(), "pos": [3, 0], "rotation": "0"} # and again
                player.fall_progress = 0.0
                lock_timer = 0.0
                player.lock_resets = 0
                state = "active"
        
        elapsed_loop = time.perf_counter() - now
        to_sleep = FRAME - elapsed_loop
        time.sleep(to_sleep if to_sleep > 0 else 0)
        
    # sigint
    inputs.stop()

def entry():
    player = Player()
    bag = Bag(preview_size=5)
    shared = { "board": setup_board(22, 10) }
    
    inputs = InputHandler()
    inputs.start()
    
    try:
        render_thread = threading.Thread(target=render_loop, args=(shared, player, bag, 60))
        render_thread.start()
        
        game_thread = threading.Thread(target=game_loop, args=(shared, player, bag, inputs, 60))
        game_thread.start()
        
    except KeyboardInterrupt:
        sigint.set()
        inputs.stop()
        game_thread.join()
        render_thread.join()

if __name__ == "__main__":    
    try:
        os.system("clear")
        entry()
    except KeyboardInterrupt:
        sys.exit(0)


# Module for configuration the user probably wouldn't see, keymaps and gravity etc

import pygame

MAX_LOCK_RESETS = 15
ARE_FRAMES = 20
LINE_CLEAR_FRAMES = 20
LOCK_DELAY_FRAMES = 30

DAS_MS = 150
ARR_MS = 50
SDF = 50

KEYMAP = {
    pygame.K_a: "move_left",
    pygame.K_d: "move_right",
    pygame.K_s: "soft_drop",
    pygame.K_w: "hard_drop",
    pygame.K_e: "hold",
    pygame.K_TAB: "rotate_180",
    pygame.K_SPACE: "rotate_cw",
    pygame.K_q: "rotate_ccw",
    pygame.K_ESCAPE: "pause",
}
""" # keymap for normal people
KEYMAP = {
    pygame.K_LEFT: "move_left",
    pygame.K_RIGHT: "move_right",
    pygame.K_DOWN: "soft_drop",
    pygame.K_SPACE: "hard_drop",
    pygame.K_LSHIFT: "hold",
    pygame.K_a: "rotate_180",
    pygame.K_UP: "rotate_cw",
    pygame.K_x: "rotate_cw",
    pygame.K_z: "rotate_ccw",
    pygame.K_ESCAPE: "pause",
}
"""

controls = {
    "move_left":  ["a", "j", "\x1b[D"], # left arrow
    "move_right": ["d", "l", "\x1b[C"], # right
    "soft_drop":  ["s", "k", "\x1b[B"], # down
    "hard_drop":  ["w", "i"],
    "rotate_cw":  ["c", " ", "\x1b[A", "/"], # up
    "rotate_ccw": ["z", "q", ","],
    "rotate_180": ["\t", "x", "."], # tab
    "hold":       ["e", "v"],
    "pause":      ["\x1b", "p"],
}


# DRAW.PY
# Module for drawing to the screen (stdout tomfuckery time!!!)
import sys
import shutil
from cetragm.tables import pieces, grades
from cetragm.game import collides

BLOCK = "██"
SHADOW = "▒▒"
EMPTY = "\x1b[48;2;20;20;20m\x1b[38;2;40;40;40m[]\x1b[49m\x1b[39m"
TOP_EMPTY = "  "

termwidth = 0
termheight = 0

def validate_board(board): # check for dimensions in case game loop fucked me
    if len(board) != 22:
        raise ValueError(f"Board has {len(board)} rows")
    widths = {len(r) for r in board}
    if widths != {10}:
        raise ValueError(f"Board has rows not 10 wide - {widths}")

def format_time(ms: int) -> str: # for TGM style timer at the bottom
    minutes = (ms // 1000) // 60 # Shame i can't make it obnoxiously big
    seconds = (ms // 1000) % 60  # or can i...? *mice on venus*
    hundredths = (ms % 1000) // 10
    return f"{minutes:02}:{seconds:02}:{hundredths:02}"

def color_block(piece_id: str | None): # in ["i", "z", "s", "l", "j", "t", "o"]
    if (not piece_id) and (not piece_id.endswith("_sh")):
        return EMPTY
    if not piece_id.endswith("_sh"):
        r, g, b = pieces[piece_id]["rgb"]
        return f"\x1b[38;2;{r};{g};{b}m{BLOCK}\x1b[0m"
    else:
        return f"\x1b[38;2;130;130;130m{SHADOW}\x1b[0m"

# public
def draw_board(board,                              # from game.py. Board: 2d list. Entries: [0] or [1, "t"]. 
               active_piece = None,                # {"name": "t", "pos": (0, 0), "rotation": "0"}
               score: int = 0, grade: str = "9",   # score is used to calculate TGM grade (9 -> 1 -> S1 -> S9 -> Gm)
               time_ms: int = 0,                   # BIG FLASHING TIMER
               lines: int = 0, line_goal: int = 0, # for gravity level up with timer
               hold: str | None = None,            # hold = "t"
               next_queue: list[str] | None = None # list of 5 pieces like ["t", "i", "z", "s", "o"] (7-bag)
               ):
    validate_board(board)
    height = len(board)
    width = len(board[0])
    
    # display on top of static board then check collision in game.py so it doesn't lag render loop
    overlay = [[cell[:] for cell in row] for row in board] # already the full board, needs to be colored
    
    # SHADOW PIECE
    if active_piece:
        pid = active_piece["name"]
        shape = pieces[pid]["rotations"][active_piece["rotation"]]
        px, py = active_piece["pos"]
        
        drop_y = py
        while True:
            drop_y += 1    
            if collides({"name": pid, "rotation": active_piece["rotation"], "pos": [px, drop_y]}, board):
                drop_y -= 1
                break
        
        for y, row in enumerate(shape):
            for x, val in enumerate(row):
                if val:
                    bx, by = px + x, drop_y + y
                    if 0 <= bx < width and 0 <= by < height:
                        overlay[by][bx] = [1, f"{pid}_sh"]
    
    # ACTUAL PIECE
    if active_piece:
        pid = active_piece["name"]
        shape = pieces[pid]["rotations"][active_piece["rotation"]]
        px, py = active_piece["pos"]
        for y, row in enumerate(shape):
            for x, val in enumerate(row):
                if val:
                    bx, by = px + x, py + y
                    if 0 <= bx < width and 0 <= by < height:
                        overlay[by][bx] = [1, pid]
                        
    left_lines = []
    left_lines.append("│ HOLD ▼ │")
    left_lines.append("╞════════╡")
    if hold and hold in pieces:
        for line in pieces[hold]["piece"]:
            left_lines.append(f"│{line.center(8)}│") # should already be centered but why not do it again
    else:
        left_lines.extend(["│   ╲╱   │", "│   ╱╲   │"])
    left_lines.append("╰────────┤")
    for i in grades[grade]:
        left_lines.append(i + "│")
        
    right_lines = []
    right_lines.append("│ NEXT ▼ │")
    right_lines.append("╞════════╡")
    
    if next_queue:
        for num, name in enumerate(next_queue[:5]):
            if name in pieces:
                for line in pieces[name]["piece"]:
                    right_lines.append(f"│{line.center(8)}│")
            else:
                right_lines.extend(["│   ╲╱   │", "│   ╱╲   │"])
            right_lines.append("├────────┤" if num != 4 else "├────────╯")
    else:
        for i in range(5):
            right_lines.extend(["│   ╲╱   │", "│   ╱╲   │", "├────────┤" if i != 4 else "├────────╯"])
    right_lines.append( "│         ")
    right_lines.append( "│ LEVEL:  ")
    right_lines.append(f"│  \x1b[4m{str(lines):^3}\x1b[24m")
    right_lines.append(f"│  {str(line_goal).ljust(9)}")
    
    frame_lines = []
    
    score = str(score) if len(str(score)) % 2 == 0 else ("0" + str(score))
    wid = (width - (len(score) // 2)) - 1
    frame_lines.append("╭────────┬"  + "─" * (wid) + f"╢{score}╟" + "─" * (wid) + "┬────────╮")
    
    for y in range(height):
        row_buf = []
        for x in range(width):
            cell = overlay[y][x]
            if y > 1:
                row_buf.append(color_block(cell[1]) if cell[0] else EMPTY)
            else:
                row_buf.append(color_block(cell[1]) if cell[0] else TOP_EMPTY)
                
        right = right_lines[y] if y < len(right_lines) else "│         "
        left = left_lines[y] if y < len(left_lines) else "         │"
        frame_lines.append(left + "".join(row_buf) + right)
    
    frame_lines.append( "         ├" + "──" * width + "┤")
    frame_lines.append(f"         │ ▶ {format_time(time_ms):^14}◀  │")
    frame_lines.append( "         ╰" + "──" * width + "╯")
    
    term = shutil.get_terminal_size()
    frame_width = len(frame_lines[0])
    pad_x = max((term.columns - frame_width) // 2, 0)
    pad_y = max((term.lines - len(frame_lines)) // 2, 0)
    
    global termwidth, termheight
    if pad_x != termwidth or pad_y != termheight:
        sys.stdout.write("\x1b[H\x1b[2J")
        termwidth, termheight = pad_x, pad_y
    
    centered = ("\n" * pad_y) + "\n".join(" " * pad_x + ln for ln in frame_lines)
    
    sys.stdout.write("\x1b[H" + centered + "\n\n")
    sys.stdout.flush()


import sys
import shutil
import time
import pygame_sdl2
import queue

from cetragm import config

MENU_COLOR_SELECTED = "\x1b;38;2;127;255;212m"
MENU_COLOR_NORMAL   = "\x1b;38;2;220;220;220m"
MENU_COLOR_VALUE    = "\x1b;38;2;13;152;186m"

def center_lines(lines):
    term = shutil.get_terminal_size()
    max_width = max(len(ln) for ln in lines) if lines else 0
    


# core logic like lock, gravity, and scoring
from cetragm.tables import pieces, ROT_180, ROT_CCW, ROT_CW
import math

def get_cells(piece): # {"name": "i", "pos": (0, 0), "rotation": "0"}
    # Gets the cells in a falling piece with its current rotation and coordinates on the board
    shape = pieces[piece["name"]]["rotations"][piece["rotation"]]
    px, py = piece["pos"]
    cells = [(px + x, py + y)
             for y, row in enumerate(shape)
             for x, val in enumerate(row) if val]
    return cells

def collides(piece, board):
    for (x, y) in get_cells(piece):
        if x < 0 or x >= len(board[0]) or y >= len(board):
            return True
        if y >= 0 and board[y][x][0]:
            return True
    return False

def lock_piece(piece, board, player):
    if collides(piece, board):
        if piece["pos"][1] == 0:
            return board, 0, True
        else:
            raise ValueError("Tried to lock outside the board or lock piece over another piece (and it wasn't a loss)")
    for i in get_cells(piece):
        board[i[1]][i[0]] = [1, piece["name"]]
        
    board, cleared = clear_lines(board)
    board_empty = all(cell[0] for row in board for cell in row)
    soft = player.soft # TODO: ADD SOFT, OR SOFT DROP CALC
    
    if board_empty:
        player.score += 2500
    
    score_gain, player.combo = get_score(player.level, cleared, player.combo, board_empty, soft)
    player.score += score_gain
    
    update_level(player, cleared)

    return board, cleared, False

def clear_lines(board):
    new_board = []
    cleared = 0
    
    for row in board:
        if all(cell[0] for cell in row):
            cleared += 1
        else:
            new_board.append(row)
            
    for _ in range(cleared):
        new_board.insert(0, [[0] for _ in range(len(board[0]))])
    
    return new_board, cleared

def get_score(level, lines_cleared, combo, board_empty, soft):
    if lines_cleared == 0:
        return 0, 1 # 0 score, reset combo
    
    combo = combo + (2 * lines_cleared) - 2 if combo > 1 else (2 * lines_cleared) - 1
    if combo <1:
        combo = 1
    bravo = 4 if board_empty else 1
    
    score = (math.ceil((level + lines_cleared) / 4) + soft) * lines_cleared * combo * bravo
    return score, combo

def update_level(player, cleared):
    if player.level >= 999:
        player.line_goal = 999
        player.level = 999
        return
    
    old_level = player.level
    temp = old_level + 1 + int(cleared)
    temp = 999 if temp > 999 else temp
        
    old_sect = old_level // 100
    new_sect = temp // 100
    
    if new_sect > old_sect and cleared == 0:
        temp = min(temp, (old_sect * 100 + 99))
        
    player.level = min(temp, 999)
    
    if player.level > 900:
        player.line_goal = 999
    else:
        player.line_goal = ((player.level // 100) + 1) * 100
        
def rotate(current, direction):
    if direction == +1:
        return ROT_CW[current]
    elif direction == -1:
        return ROT_CCW[current]
    elif direction == 2:
        return ROT_180[current]
    raise ValueError("direction should be -1, +1, or 2.")
    return current


# Module for binding and rebinding controls as well as taking input (threaded)
# DAS/ARR logic

from cetragm.config import controls
import termios
import tty
import os
import threading
import queue
import select
import sys

class InputHandler:
    def __init__(self):
        self.queue = queue.Queue()
        self.stop_flag = threading.Event()
        self.is_windows = os.name == "nt"
        self.fd = None
        self.old_settings = None
            
    def start(self):
        self.thread = threading.Thread(target=self._poll_loop, daemon=True)
        self.thread.start()
    
    def stop(self):
        self.stop_flag.set()
        if self.is_windows:
            return
        if self.fd is not None and self.old_settings is not None:
            try:
                termios.tcsetattr(self.fd, termios.TCSADRAIN, self.old_settings)
            except termios.error:
                pass
    
    def _poll_loop(self):
        if self.is_windows:
            self._poll_windows()
        else:
            self._poll_linux()
    
    def _poll_windows(self):
        import msvcrt
        while not self.stop_flag.is_set():
            if msvcrt.kbhit():
                ch = msvcrt.getwch()
                self._process_key(ch)
            else:
                threading.Event().wait(0.01)
    
    def _poll_linux(self):
        try:
            self.fd = os.open("/dev/tty", os.O_RDONLY)
        except OSError:
            self.fd = sys.stdin.fileno()
            
        if not os.isatty(self.fd):
            print("not a tty!! failing out", file=sys.stderr)
            return
        
        self.old_settings = termios.tcgetattr(self.fd)
        tty.setcbreak(self.fd)
        
        try:
            while not self.stop_flag.is_set():
                r, _, _ = select.select([self.fd], [], [], 0.02)
                if not r:
                    continue
                
                try:
                    ch = os.read(self.fd, 1).decode(errors="ignore")
                except OSError as e:
                    print("OSerror:", e, file=sys.stderr)
                    break
                
                if not ch:
                    continue
                
                if ch == "\x1b":
                    seq = ch
                    while select.select([self.fd], [], [], 0.001)[0]:
                        seq += os.read(self.fd, 1).decode(errors="ignore")
                    key = seq
                else:
                    key = ch
                self._process_key(key)
        finally:
            if self.old_settings is not None and os.isatty(self.fd):
                try:
                    termios.tcsetattr(self.fd, termios.TCSADRAIN, self.old_settings)
                except termios.error:
                    pass
            if self.fd not in (None, sys.stdin.fileno()):
                try:
                    os.close(self.fd)
                except OSError as e:
                    print(e, file=sys.stderr)
          
    def _process_key(self, key):
        for action, binds in controls.items():
            if key in binds:
                self.queue.put(action)
                break


# 7-bag randomizer

import random
from cetragm.tables import pieces

PIECE_BAG = list(pieces.keys())

class Bag:
    def __init__(self, preview_size: int = 5):
        self.preview_size = preview_size
        self.current_bag = random.sample(PIECE_BAG, len(PIECE_BAG))
        self.next_bag = random.sample(PIECE_BAG, len(PIECE_BAG))
    
    def get_piece(self):
        piece = self.current_bag.pop(0)
        if not self.current_bag:
            self.current_bag = self.next_bag
            self.next_bag = random.sample(PIECE_BAG, len(PIECE_BAG))
        return piece
    
    def get_preview(self):
        upcoming = self.current_bag + self.next_bag
        return upcoming[:self.preview_size]
    


# das/arr with actual keydown *requires a pygame window

import pygame
import threading
import queue
import time
import sys
import os

from cetragm.config import DAS_MS, ARR_MS, SDF, KEYMAP

class InputHandler:
    def __init__(self, keymap=None, window_size = (200, 100), hidden = True):
        self.queue = queue.Queue()
        self._thread = None
        self._running = False
        self._pressed = set() # keys to ARR
        self._repeat_state = {} # wait for das or repeat until x
        self.keymap = keymap or KEYMAP
        self.window_size = window_size
        self.hidden = hidden
        
        self.DAS = DAS_MS / 1000.0
        self.ARR = ARR_MS / 1000.0
        self.SDF = SDF / 1000.0
        
        self._movement_keys = {k for k, v in self.keymap.items() if v in ("move_left", "move_right")} # what to apply DAS to
        
    def start(self):
        if self._running:
            return
        self._running = True
        self._thread = threading.Thread(target=self._run, daemon=True)
        self._thread.start()
    
    def stop(self):
        self._running = False
        try:
            pygame.event.post(pygame.event.Event(pygame.QUIT)) # try to exit gracefully
        except Exception:
            pass
        
        if self._thread is not None: # if that doesn't work bonk it
            self._thread.join(timeout= 0.25)
    
    def movement_pressed(self):
        return any(k in self._pressed for k in self._movement_keys)
    
    def _enqueue(self, action):
        try:
            self.queue.put_nowait(action)
        except Exception:
            pass
        
    def _run(self):
        os.environ['SDL_VIDEO_WINDOW_POS'] = "%d,%d" % (30, 30)
        
        pygame.init()
        flags = 0   
        if self.hidden:
            flags = pygame.NOFRAME
        try:
            screen = pygame.display.set_mode(self.window_size, flags)
        except Exception:
            screen = pygame.display.set_mode((320, 64))
        pygame.display.set_caption("Focus for CGM input")
        
        try: # attempt to auto-grab input, may not work on hyprland etc
            pygame.event.set_blocked(None)
            pygame.event.set_allowed([pygame.KEYDOWN, pygame.KEYUP, pygame.QUIT])
            pygame.event.set_keyboard_grab(True)
            pygame.key.set_repeat() # turn off pygame native ARR
        except Exception:
            pass
        
        font = None
        try:
            font = pygame.font.Font(None, 18)
        except Exception:
            pass
        
        def _draw_message():
            screen.fill((0, 100, 80))
            lines = [
                "this window gets input for CGM!",
                "can't do it via terminal because",
                "there's no press/release input.",
                "press ESC to close this window,",
                "and focus it to play!",
                "-@qwik (ps rate 5 stars)"
            ]
            if font:
                y = 6
                for ln in lines:
                    surf = font.render(ln, True, (220, 220, 220))
                    screen.blit(surf, (8, y))
                    y += surf.get_height() + 2
            else:
                pass
            pygame.display.flip()
        
        _draw_message()
        
        last_time = time.monotonic()
        while self._running:
            now = time.monotonic()
            dt = now - last_time
            last_time = now
            
            for ev in pygame.event.get():
                if ev.type == pygame.QUIT:
                    self._enqueue("pause")
                    self._running = False
                    break
                if ev.type == pygame.KEYDOWN:
                    key = ev.key
                    was_down = key in self._pressed
                    self._pressed.add(key)
                    
                    self._repeat_state[key] = {
                        "das_until": now + self.DAS,
                        "next_repeat": now + self.DAS + self.ARR
                    }
                    
                    action = self.keymap.get(key)
                    if action is None:
                        continue
                    
                    self._enqueue(action)
                
                elif ev.type == pygame.KEYUP:
                    key = ev.key
                    if key in self._pressed:
                        self._pressed.remove(key)
                    if key in self._repeat_state:
                        del self._repeat_state[key]
                        
            now = time.monotonic()
            for key, st in list(self._repeat_state.items()):
                action = self.keymap.get(key)
                if action not in ("move_left", "move_right"):
                    continue
                
                if now >= st["das_until"]:
                    if now >= st["next_repeat"]:
                        self._enqueue(action)
                        st["next_repeat"] = now + self.ARR
                        self._repeat_state[key] = st
                        
            sd_keys = [k for k in self._pressed if self.keymap.get(k) == "soft_drop"]
            if sd_keys:
                key = "__soft_drop__"
                st = self._repeat_state.get(key)
                
                if st is None:
                    self._enqueue("soft_drop")
                    self._repeat_state[key] = {
                        "next_repeat": now + self.SDF
                    }
                else:
                    if now >= st["next_repeat"]:
                        self._enqueue("soft_drop")
                        st["next_repeat"] = now + self.SDF
            else:
                self._repeat_state.pop("__soft_drop__", None)
                    
            time.sleep(0.01)
        
        try:
            pygame.event.set_keyboard_grab(False)
        except Exception:
            pass
        try:
            pygame.quit()
        except Exception:
            pass


# Structure:
# main.py
# |- audio.py    - sfx, music
# |- controls.py - das, arr
# |- draw.py     - ui, stdout, terminal
# |- srs.py      - wallkicks and rotation
# |
# |- game.py     - scoring, logic, hold, etc
# |- pieces.py   - define tetrominoes
# |- config.py   - rebind controls, gravity, etc
# |
# \- /assets     - for audio.py (simpleaudio)
#    |- /sfx
#    \- /music

pass